@@ -75,6 +75,8 @@ typedef struct {
 	SoupMessageQueue *queue;
 
 	char *user_agent;
+	char *accept_language;
+	gboolean accept_language_auto;
 
 	GSList *features;
 	GHashTable *features_cache;
@@ -141,6 +143,8 @@ enum {
 	PROP_ASYNC_CONTEXT,
 	PROP_TIMEOUT,
 	PROP_USER_AGENT,
+	PROP_ACCEPT_LANGUAGE,
+	PROP_ACCEPT_LANGUAGE_AUTO,
 	PROP_IDLE_TIMEOUT,
 	PROP_ADD_FEATURE,
 	PROP_ADD_FEATURE_BY_TYPE,
@@ -216,6 +220,7 @@ finalize (GObject *object)
 	g_hash_table_destroy (priv->conns);
 
 	g_free (priv->user_agent);
+	g_free (priv->accept_language);
 
 	if (priv->ssl_creds)
 		soup_ssl_free_client_credentials (priv->ssl_creds);
@@ -390,7 +395,10 @@ soup_session_class_init (SoupSessionClass *session_class)
 			      NULL, NULL,
 			      soup_marshal_NONE__OBJECT,
 			      G_TYPE_NONE, 1,
-			      SOUP_TYPE_CONNECTION);
+			      /* SoupConnection is private, so we can't use
+			       * SOUP_TYPE_CONNECTION here.
+			       */
+			      G_TYPE_OBJECT);
 
 	signals[TUNNELING] =
 		g_signal_new ("tunneling",
@@ -400,7 +408,10 @@ soup_session_class_init (SoupSessionClass *session_class)
 			      NULL, NULL,
 			      soup_marshal_NONE__OBJECT,
 			      G_TYPE_NONE, 1,
-			      SOUP_TYPE_CONNECTION);
+			      /* SoupConnection is private, so we can't use
+			       * SOUP_TYPE_CONNECTION here.
+			       */
+			      G_TYPE_OBJECT);
 
 
 	/* properties */
@@ -562,6 +573,55 @@ soup_session_class_init (SoupSessionClass *session_class)
 				     NULL,
 				     G_PARAM_READWRITE));
 
+	/**
+	 * SoupSession:accept-language:
+	 *
+	 * If non-%NULL, the value to use for the "Accept-Language" header
+	 * on #SoupMessage<!-- -->s sent from this session.
+	 *
+	 * Setting this will disable
+	 * #SoupSession:accept-language-auto.
+	 *
+	 * Since: 2.30
+	 **/
+	/**
+	 * SOUP_SESSION_ACCEPT_LANGUAGE:
+	 *
+	 * Alias for the #SoupSession:accept-language property, qv.
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_ACCEPT_LANGUAGE,
+		g_param_spec_string (SOUP_SESSION_ACCEPT_LANGUAGE,
+				     "Accept-Language string",
+				     "Accept-Language string",
+				     NULL,
+				     G_PARAM_READWRITE));
+
+	/**
+	 * SoupSession:accept-language-auto:
+	 *
+	 * If %TRUE, #SoupSession will automatically set the string
+	 * for the "Accept-Language" header on every #SoupMessage
+	 * sent, based on the return value of g_get_language_names().
+	 *
+	 * Setting this will override any previous value of
+	 * #SoupSession:accept-language.
+	 *
+	 * Since: 2.30
+	 **/
+	/**
+	 * SOUP_SESSION_ACCEPT_LANGUAGE_AUTO:
+	 *
+	 * Alias for the #SoupSession:accept-language-auto property, qv.
+	 **/
+	g_object_class_install_property (
+		object_class, PROP_ACCEPT_LANGUAGE_AUTO,
+		g_param_spec_boolean (SOUP_SESSION_ACCEPT_LANGUAGE_AUTO,
+				      "Accept-Language automatic mode",
+				      "Accept-Language automatic mode",
+				      FALSE,
+				      G_PARAM_READWRITE));
+
 	/**
 	 * SoupSession:add-feature:
 	 *
@@ -646,6 +706,88 @@ safe_str_equal (const char *a, const char *b)
 	return strcmp (a, b) == 0;
 }
 
+/* Converts a language in POSIX format and to be RFC2616 compliant    */
+/* Based on code from epiphany-webkit (ephy_langs_append_languages()) */
+static gchar *
+posix_lang_to_rfc2616 (const gchar *language)
+{
+	/* Don't include charset variants, etc */
+	if (strchr (language, '.') || strchr (language, '@'))
+		return NULL;
+
+	/* Ignore "C" locale, which g_get_language_names() always
+	 * includes as a fallback.
+	 */
+	if (!strcmp (language, "C"))
+		return NULL;
+
+	return g_strdelimit (g_ascii_strdown (language, -1), "_", '-');
+}
+
+/* Converts @quality from 0-100 to 0.0-1.0 and appends to @str */
+static gchar *
+add_quality_value (const gchar *str, int quality)
+{
+	g_return_val_if_fail (str != NULL, NULL);
+
+	if (quality > 0 && quality < 100) {
+		double qvalue = quality / 100.0;
+		return g_strdup_printf ("%s;q=%.2g", str, qvalue);
+	} else {
+		/* Just dup the string in this case */
+		return g_strdup (str);
+	}
+}
+
+/* Returns a RFC2616 compliant languages list from system locales */
+static gchar *
+accept_languages_from_system (void)
+{
+	const char * const * lang_names;
+	GPtrArray *langs = NULL;
+	char *lang, **langs_array, *langs_str;
+	int delta;
+	int i;
+
+	lang_names = g_get_language_names ();
+	g_return_val_if_fail (lang_names != NULL, NULL);
+
+	/* Build the array of languages */
+	langs = g_ptr_array_new ();
+	for (i = 0; lang_names[i] != NULL; i++) {
+		lang = posix_lang_to_rfc2616 (lang_names[i]);
+		if (lang)
+			g_ptr_array_add (langs, lang);
+	}
+
+	/* Add quality values */
+	if (langs->len < 10)
+		delta = 10;
+	else if (langs->len < 20)
+		delta = 5;
+	else
+		delta = 1;
+
+	for (i = 0; i < langs->len; i++) {
+		lang = langs->pdata[i];
+		langs->pdata[i] = add_quality_value (lang, 100 - i * delta);
+		g_free (lang);
+	}
+
+	/* Fallback: add "en" if list is empty */
+	if (langs->len == 0)
+		g_ptr_array_add (langs, g_strdup ("en"));
+
+	g_ptr_array_add (langs, NULL);
+	langs_array = (char **)langs->pdata;
+	langs_str = g_strjoinv (", ", langs_array);
+
+	g_strfreev (langs_array);
+	g_ptr_array_free (langs, FALSE);
+
+	return langs_str;
+}
+
 static void
 set_property (GObject *object, guint prop_id,
 	      const GValue *value, GParamSpec *pspec)
@@ -724,6 +866,22 @@ set_property (GObject *object, guint prop_id,
 		} else
 			priv->user_agent = g_strdup (user_agent);
 		break;
+	case PROP_ACCEPT_LANGUAGE:
+		g_free (priv->accept_language);
+		priv->accept_language = g_strdup (g_value_get_string (value));
+		priv->accept_language_auto = FALSE;
+		break;
+	case PROP_ACCEPT_LANGUAGE_AUTO:
+		priv->accept_language_auto = g_value_get_boolean (value);
+		if (priv->accept_language) {
+			g_free (priv->accept_language);
+			priv->accept_language = NULL;
+		}
+
+		/* Get languages from system if needed */
+		if (priv->accept_language_auto)
+			priv->accept_language = accept_languages_from_system ();
+		break;
 	case PROP_IDLE_TIMEOUT:
 		priv->idle_timeout = g_value_get_uint (value);
 		break;
@@ -787,6 +945,12 @@ get_property (GObject *object, guint prop_id,
 	case PROP_USER_AGENT:
 		g_value_set_string (value, priv->user_agent);
 		break;
+	case PROP_ACCEPT_LANGUAGE:
+		g_value_set_string (value, priv->accept_language);
+		break;
+	case PROP_ACCEPT_LANGUAGE_AUTO:
+		g_value_set_boolean (value, priv->accept_language_auto);
+		break;
 	case PROP_IDLE_TIMEOUT:
 		g_value_set_uint (value, priv->idle_timeout);
 		break;
@@ -985,6 +1149,14 @@ soup_session_send_queue_item (SoupSession *session,
 					      "User-Agent", priv->user_agent);
 	}
 
+	if (priv->accept_language &&
+	    !soup_message_headers_get_list (item->msg->request_headers,
+					    "Accept-Language")) {
+		soup_message_headers_append (item->msg->request_headers,
+					     "Accept-Language",
+					     priv->accept_language);
+	}
+
 	g_signal_emit (session, signals[REQUEST_STARTED], 0,
 		       item->msg, soup_connection_get_socket (conn));
 	soup_connection_send_request (conn, item->msg);
@@ -1058,10 +1230,14 @@ soup_session_connection_failed (SoupSession *session,
 	g_mutex_lock (priv->host_lock);
 	host = g_hash_table_lookup (priv->conns, conn);
 	g_mutex_unlock (priv->host_lock);
-	if (!host)
-		return;
 
-	connection_disconnected (conn, session);
+	if (host)
+		connection_disconnected (conn, session);
+	else {
+		host = g_object_get_data (G_OBJECT (conn), "SoupSessionHost");
+		if (!host)
+			return;
+	}
 
 	/* Cancel any other messages waiting for a connection to it,
 	 * since they're out of luck.
@@ -1236,6 +1412,7 @@ soup_session_get_connection (SoupSession *session,
 		SOUP_CONNECTION_TIMEOUT, priv->io_timeout,
 		SOUP_CONNECTION_IDLE_TIMEOUT, priv->idle_timeout,
 		NULL);
+	g_object_set_data (G_OBJECT (conn), "SoupSessionHost", host);
 	g_signal_connect (conn, "disconnected",
 			  G_CALLBACK (connection_disconnected),
 			  session);
